import pytest

from core.util import json_util


class TestJsonUtil:

    def test_dumps_with_none(self):
        assert json_util.dumps(None) == 'null'

    def test_dumps_with_boolean_true(self):
        assert json_util.dumps(True) == 'true'

    def test_dumps_with_boolean_false(self):
        assert json_util.dumps(False) == 'false'

    def test_dumps_with_integer(self):
        assert json_util.dumps(42) == '42'

    def test_dumps_with_float(self):
        assert json_util.dumps(3.14) == '3.14'

    def test_dumps_with_string(self):
        assert json_util.dumps("hello") == '"hello"'

    def test_dumps_with_list(self):
        assert json_util.dumps([1, 2, 3]) == '[1,2,3]'

    def test_dumps_with_dict(self):
        assert json_util.dumps({"key": "value"}) == '{"key":"value"}'

    def test_dumps_with_nested_dict(self):
        data = {
            "name": "test",
            "nested": {
                "a": True,
                "b": [4, 5, 6],
            }
        }
        result = json_util.dumps(data)
        assert json_util.loads(result) == data

    def test_dumps_with_newline(self):
        assert json_util.dumps("Hello\nWorld") == '"Hello\\nWorld"'

    def test_dumps_with_tab(self):
        assert json_util.dumps("Hello\tWorld") == '"Hello\\tWorld"'

    def test_dumps_with_unicode_emoji(self):
        assert json_util.dumps("Hello 🌍") == '"Hello 🌍"'

    def test_dumps_with_quotes(self):
        assert json_util.dumps('Hello "World"') == '"Hello \\"World\\""'

    def test_dumps_with_custom_class_raises_error(self):
        class TestClass:
            pass
        with pytest.raises(json_util.JsonEncodeException) as exceptionInfo:
            json_util.dumps(TestClass())
        assert exceptionInfo.value.message == "Type is not JSON serializable: TestClass"

    def test_dumps_with_complex_number_raises_error(self):
        with pytest.raises(json_util.JsonEncodeException):
            json_util.dumps(complex(1, 2))

    def test_loads_with_null(self):
        assert json_util.loads('null') is None

    def test_loads_with_boolean_true(self):
        assert json_util.loads('true') is True

    def test_loads_with_boolean_false(self):
        assert json_util.loads('false') is False

    def test_loads_with_integer(self):
        assert json_util.loads('42') == 42

    def test_loads_with_float(self):
        assert json_util.loads('3.14') == 3.14

    def test_loads_with_string(self):
        assert json_util.loads('"hello"') == "hello"

    def test_loads_with_list(self):
        assert json_util.loads('[1,2,3]') == [1, 2, 3]

    def test_loads_with_dict(self):
        assert json_util.loads('{"key":"value"}') == {"key": "value"}

    def test_loads_with_nested_dict(self):
        jsonStr = '{"name":"test","nested":{"a":true,"b":[4,5,6]}}'
        expected = {
            "name": "test",
            "nested": {
                "a": True,
                "b": [4, 5, 6],
            }
        }
        assert json_util.loads(jsonStr) == expected

    def test_loads_with_newline(self):
        assert json_util.loads('"Hello\\nWorld"') == "Hello\nWorld"

    def test_loads_with_tab(self):
        assert json_util.loads('"Hello\\tWorld"') == "Hello\tWorld"

    def test_loads_with_unicode_emoji(self):
        assert json_util.loads('"Hello 🌍"') == "Hello 🌍"

    def test_loads_with_quotes(self):
        assert json_util.loads('"Hello \\"World\\""') == 'Hello "World"'

    def test_loads_with_incomplete_object_raises_error(self):
        with pytest.raises(json_util.JsonDecodeException):
            json_util.loads('{')

    def test_loads_with_trailing_comma_raises_error(self):
        with pytest.raises(json_util.JsonDecodeException):
            json_util.loads('[1, 2,]')

    def test_loads_with_unquoted_string_raises_error(self):
        with pytest.raises(json_util.JsonDecodeException):
            json_util.loads('{"key": value}')

    def test_loads_with_undefined_raises_error(self):
        with pytest.raises(json_util.JsonDecodeException):
            json_util.loads('{"key": undefined}')

    def test_loads_with_single_quotes_raises_error(self):
        with pytest.raises(json_util.JsonDecodeException):
            json_util.loads("{'key': 'value'}")

    def test_loads_with_str_input(self):
        assert json_util.loads('{"key":"value"}') == {"key": "value"}

    def test_loads_with_bytes_input(self):
        assert json_util.loads(b'{"key":"value"}') == {"key": "value"}

    def test_loads_with_bytearray_input(self):
        assert json_util.loads(bytearray(b'{"key":"value"}')) == {"key": "value"}

    def test_loads_with_whitespace(self):
        jsonStr = '''
        {
            "a": 1,
            "b": 2
        }
        '''
        assert json_util.loads(jsonStr) == {"a": 1, "b": 2}

    def test_dumpb_with_very_long_integer(self):
        veryLongInteger = 11111111111111111111111
        result = json_util.dumpb(obj=veryLongInteger)
        assert isinstance(result, bytes)
        assert result == b'11111111111111111111111'

    # NOTE(krishan711): not sure why this is failing
    # def test_dumpb_and_loads_with_very_long_integer(self):
    #     veryLongInteger = 11111111111111111111111
    #     result = json_util.dumpb(obj=veryLongInteger)
    #     assert isinstance(result, bytes)
    #     assert int(json_util.loads(result)) == veryLongInteger



from collections.abc import Mapping
from collections.abc import Sequence

JsonBaseType = str | int | float | bool | None
type Json = dict[str, 'Json'] | Mapping[str, 'Json'] | list['Json'] | Sequence['Json'] | JsonBaseType
type JsonObject = dict[str, 'Json']
type JsonList = list['Json']



import typing

from starlette.responses import JSONResponse

from core.util import json_util


class KibaJSONResponse(JSONResponse):
    def render(self, content: typing.Any) -> bytes:  # type: ignore[explicit-any]
        return json_util.dumpb(content)



import pytest
from starlette.applications import Starlette
from starlette.routing import Route
from starlette.testclient import TestClient
from pydantic import BaseModel

from core.api.json_route import json_route
from core.api.api_request import KibaApiRequest
from core.api.middleware.exception_handling_middleware import ExceptionHandlingMiddleware


class ExampleRequest(BaseModel):
    name: str
    age: int
    tags: list[str] | None = None
    path_param: str | None = None


class ExampleResponse(BaseModel):
    message: str
    request_name: str
    request_age: int
    path_echo: str | None = None


@pytest.fixture
def client():
    @json_route(requestType=ExampleRequest, responseType=ExampleResponse)
    async def test_endpoint(request: KibaApiRequest[ExampleRequest]) -> ExampleResponse:
        return ExampleResponse(
            message="Hello, world!",
            request_name=request.data.name,
            request_age=request.data.age,
            path_echo=request.data.path_param,
        )


    @json_route(requestType=ExampleRequest, responseType=ExampleResponse)
    async def test_path_endpoint(request: KibaApiRequest[ExampleRequest]) -> ExampleResponse:
        return ExampleResponse(
            message="Path parameter received",
            request_name=request.data.name,
            request_age=request.data.age,
            path_echo=request.data.path_param,
        )


    app = Starlette(routes=[
        Route("/test", test_endpoint, methods=["POST"]),
        Route("/test/{path_param}", test_path_endpoint, methods=["POST"]),
    ])
    app.add_middleware(ExceptionHandlingMiddleware)
    return TestClient(app, raise_server_exceptions=False)


def test_json_route_with_valid_body(client):
    response = client.post(
        "/test",
        json={"name": "Test User", "age": 30, "tags": ["tag1", "tag2"]}
    )
    assert response.status_code == 200
    data = response.json()
    assert data["message"] == "Hello, world!"
    assert data["request_name"] == "Test User"
    assert data["request_age"] == 30


def test_json_route_with_path_params(client):
    response = client.post(
        "/test/param-value",
        json={"name": "Test User", "age": 30}
    )
    assert response.status_code == 200
    data = response.json()
    assert data["path_echo"] == "param-value"
    assert data["request_name"] == "Test User"


def test_json_route_with_query_params(client):
    response = client.post(
        "/test?name=QueryUser&age=25",
        json={}
    )
    assert response.status_code == 200
    data = response.json()
    assert data["request_name"] == "QueryUser"
    assert data["request_age"] == 25


def test_json_route_with_invalid_json(client):
    response = client.post(
        "/test",
        content="invalid json",
        headers={"Content-Type": "application/json"}
    )
    assert response.status_code == 400
    assert "Invalid JSON body" in response.json().get("message", "")


def test_json_route_with_validation_error_missing_field(client):
    response = client.post(
        "/test",
        json={"name": "Test User"}
    )
    assert response.status_code == 400
    assert "Invalid request" in response.json().get("message", "")


def test_json_route_combining_params(client):
    response = client.post(
        "/test/path-value?age=40",
        json={"name": "Combined User"}
    )
    assert response.status_code == 200
    data = response.json()
    assert data["path_echo"] == "path-value"
    assert data["request_name"] == "Combined User"
    assert data["request_age"] == 40


def test_json_route_empty_body(client):
    response = client.post(
        "/test?name=EmptyBody&age=50",
    )
    assert response.status_code == 200
    data = response.json()
    assert data["request_name"] == "EmptyBody"
    assert data["request_age"] == 50



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response

from core.exceptions import KibaException
from core.exceptions import RedirectException


class ExceptionHandlingMiddleware(BaseHTTPMiddleware):
    @staticmethod
    def _convert_exception(exception: KibaException) -> Response:
        response = JSONResponse(status_code=exception.statusCode, content=exception.to_dict())
        return response

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        try:
            response = await call_next(request)
        except RedirectException as exception:
            response = Response(status_code=exception.statusCode)
            response.headers['Location'] = exception.location
            # TODO(krishan711): clean this up after major version bump (since shouldAddCacheHeader field will always exist)
            if not hasattr(exception, 'shouldAddCacheHeader') or exception.shouldAddCacheHeader:
                response.headers['Cache-Control'] = f'max-age={60 * 60 * 24 * 365}'
        except KibaException as exception:
            # logging.exception(exception)
            response = self._convert_exception(exception=exception)
        except Exception as exception:  # noqa: BLE001
            # logging.exception(exception)
            response = self._convert_exception(exception=KibaException.from_exception(exception=exception))
        return response



import datetime
import json
import typing
from typing import Any

import orjson

from core import logging
from core.exceptions import KibaException
from core.util import date_util
from core.util.typing_util import Json


class JsonDecodeException(KibaException):
    pass


class JsonEncodeException(KibaException):
    pass


_HAS_LOGGED_FOR_SERIALIZATION_ERROR = False


class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj: typing.Any) -> str:  # type: ignore[explicit-any]
        if isinstance(obj, datetime.datetime):
            return date_util.datetime_to_string(obj)
        try:
            return typing.cast(str, super().default(obj))
        except TypeError:
            return str(obj)


def dumpb(obj: Any) -> bytes:  # type: ignore[explicit-any]
    global _HAS_LOGGED_FOR_SERIALIZATION_ERROR  # noqa: PLW0603
    try:
        return orjson.dumps(obj)
    except TypeError as exception:
        if str(exception) == 'Integer exceeds 64-bit range':
            if not _HAS_LOGGED_FOR_SERIALIZATION_ERROR:
                logging.warning(f'There was an error during the serialization an object: `{exception}`, falling back to json.')
                _HAS_LOGGED_FOR_SERIALIZATION_ERROR = True
            return json.dumps(obj, cls=DatetimeEncoder).encode('utf-8')
        raise JsonEncodeException(message=str(exception)) from exception
    except orjson.JSONEncodeError as exception:
        raise JsonEncodeException(message=str(exception)) from exception


def dumps(obj: Any) -> str:  # type: ignore[explicit-any]
    return dumpb(obj=obj).decode('utf-8')


def loads(json: str | bytes | bytearray) -> Json:
    try:
        return orjson.loads(json)  # type: ignore[no-any-return]
    except orjson.JSONDecodeError as exception:
        raise JsonDecodeException(message=str(exception)) from exception
